package info.textgrid.middleware.test.online;

import static org.junit.Assert.assertTrue;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import javax.ws.rs.core.Response;
import org.apache.cxf.helpers.IOUtils;
import org.apache.cxf.jaxrs.client.Client;
import org.apache.cxf.jaxrs.client.JAXRSClientFactory;
import org.apache.cxf.jaxrs.client.WebClient;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;
import org.apache.http.HttpStatus;
import info.textgrid.middleware.OAIPMHProducer;

/**
 * <p>
 * Some online tests for the TextGrid OAIMPH service.
 * </p>
 * 
 * @author Stefan E. Funk, SUB Göttingen
 */
public class OAIPMHUtilitiesOnline {

  // **
  // FINALS
  // **

  // ## CHANGE SETTINGS BELOW FOR SETTING TEST SCOPE ---------------------------------------------

  public static final String PROPERTIES_FILE = "oaipmh.test.textgridlab-org.properties";
  // public static final String PROPERTIES_FILE = "oaipmh.test.dev-textgridlab-org.properties";

  public static final boolean TEST_ALL_PAGES = false;
  // public static final boolean TEST_ALL_PAGES = true;

  // ## CHANGE SETTINGS ABOVE FOR SETTING TEST SCOPE ---------------------------------------------

  public static final String OAI_DC_PREFIX = "oai_dc";
  public static final String OAI_DATACITE_PREFIX = "oai_datacite";
  public static final String OAI_IDIOMMETS_PREFIX = "oai_idiom_mets";

  public static final String VERB_LIST_IDENTIFIERS = "ListIdentifiers";
  public static final String VERB_LIST_RECORDS = "ListRecords";
  public static final String VERB_GET_RECORD = "GetRecord";
  public static final String VERB_IDENTIFY = "Identify";
  public static final String VERB_LIST_METADATA_FORMATS = "ListMetadataFormats";
  public static final String VERB_LIST_SETS = "ListSets";

  public static final String HEADER_ID_START_TAG = "<identifier>";
  public static final String HEADER_ID_END_TAG = "</identifier>";

  public static final String EXPECTED_OAIDC_FORMAT_CONTENT = "<oai_dc:dc>";
  public static final String EXPECTED_IDIOMMETS_FORMAT_CONTENT = "<mets ";
  public static final String EXPECTED_DATACITE_FORMAT_CONTENT = "<datacite:resource>";

  public static final String OK = ">>> OKIDOKI";
  public static final String TESTING = "\n >>> TESTING ";
  public static final String NOT_TESTED = " >>> NOT_TESTED";
  public static final String NO_TOKEN = "-1";

  public static final String NO_SET = null;
  public static final String NO_THREAD_NAME = "";
  public static final boolean NO_METADATA_FORMAT_WITH_RESTOK = false;
  public static final boolean METADATA_FORMAT_WITH_RESTOK = true;
  public static final String NO_FROM = null;
  public static final String NO_UNTIL = null;
  public static final String NO_QUERY = "";

  // Time to wait between the single queries using resumption tokens in milliseconds.
  public static final Long TIME = 275l;

  public static final String HDL_PREFIX = "hdl:";
  public static final String STAR_PREFIX = "***";
  public static final String OAI_PATH = "oai";
  public static final String VERSION_PATH = "oai/version";

  public static final long OAIPMH_CLIENT_TIMEOUT = 120000;

  /**
   * @param theClient
   * @param theQuery
   * @return
   * @throws IOException
   */
  public static Response getOAIHttpResponse(Client theClient, String theQuery)
      throws IOException {
    return getHttpResponse(theClient, NO_THREAD_NAME, OAI_PATH, theQuery);
  }

  /**
   * @param theClient
   * @param theThreadName
   * @param theQuery
   * @return
   * @throws IOException
   */
  public static Response getOAIHttpResponse(Client theClient, String theThreadName,
      String theQuery) throws IOException {
    return getHttpResponse(theClient, theThreadName, OAI_PATH, theQuery);
  }

  /**
   * @param theClient
   * @return
   * @throws IOException
   */
  public static Response getVersionHttpResponse(Client theClient) throws IOException {
    return getHttpResponse(theClient, NO_THREAD_NAME, VERSION_PATH, NO_QUERY);
  }

  /**
   * @param theClient
   * @param theThreadName
   * @param thePath
   * @param theQuery
   * @return
   * @throws IOException
   */
  protected static Response getHttpResponse(Client theClient, String theThreadName, String thePath,
      String theQuery) throws IOException {

    Response result;

    try {
      Thread.sleep(TIME);
    } catch (InterruptedException e) {
      // Nothing interrupts here!
    }

    WebClient w = WebClient.fromClient(theClient).path(thePath).replaceQuery(theQuery);

    System.out.println("\t" + theThreadName + "HTTP GET " + w.getCurrentURI());

    result = w.get();

    // Check HTTP status.
    int status = result.getStatus();

    if (status != HttpStatus.SC_OK) {
      String message = theThreadName + "status: " + status;
      assertTrue(message, false);
    } else {
      System.out.println("\t" + theThreadName + "status: " + status);
    }

    return result;
  }

  /**
   * @param theMillis
   * @return
   */
  public static String getDurationInSecs(long theMillis) {

    int SECS_IN_MILLIS = 1000;
    int secs = SECS_IN_MILLIS;

    return theMillis / secs + " second" + ((theMillis / secs) != 1 ? "s" : "");
  }

  /**
   * <p>
   * Loads a resource.
   * </p>
   * 
   * TODO Put together with the method in TGCrudServiceUtilities! Maybe create a first build maven
   * module for utility things.
   * 
   * @param {@link String} The resource to search for.
   * @return {@link File} The resource.
   * @throws IOException
   */
  public static File getResource(String resPart) throws IOException {

    File res;

    // If we have an absolute resPart, just return the file.
    if (resPart.startsWith(File.separator)) {
      return new File(resPart);
    }

    URL url = ClassLoader.getSystemClassLoader().getResource(resPart);
    if (url == null) {
      throw new IOException("Resource '" + resPart + "' not found");
    }
    try {
      res = new File(url.toURI());
    } catch (URISyntaxException ue) {
      res = new File(url.getPath());
    }

    return res;
  }

  /**
   * @param theProperty
   * @return
   */
  public static List<String> getListFromProperties(String theProperty) {

    List<String> result = new ArrayList<String>();

    String parts[] = theProperty.split(",");
    for (String part : parts) {
      result.add(part.trim());
    }

    return result;
  }

  /**
   * @return
   * @throws IOException
   * @throws FileNotFoundException
   */
  public static Properties getPropertiesFromFile(String theConfigFile)
      throws FileNotFoundException, IOException {

    Properties result = null;

    // Load properties file.
    result = new Properties();
    result.load(new FileInputStream(getResource(theConfigFile)));

    System.out.println("Properties file: " + theConfigFile);
    result.list(System.out);

    return result;
  }

  /**
   * @param theEndpoint
   * @return
   */
  public static Client getOAIPMHWEebClient(String theEndpoint) {

    Client result = null;

    // Get OAI-PMH REST endpoint and HTTP client.
    System.out.println("Getting OAI-PMH HTTP client --> " + theEndpoint + " <--");

    // Get proxy first, set policy.
    OAIPMHProducer JAXRSClient = JAXRSClientFactory.create(theEndpoint, OAIPMHProducer.class);
    HTTPConduit conduit = WebClient.getConfig(JAXRSClient).getHttpConduit();
    HTTPClientPolicy policy = new HTTPClientPolicy();
    policy.setReceiveTimeout(OAIPMHUtilitiesOnline.OAIPMH_CLIENT_TIMEOUT);
    conduit.setClient(policy);

    // Create Web Client from Web Proxy.
    result = WebClient.client(JAXRSClient);

    return result;
  }

  // **
  // TG used methods
  // **

  /**
   * @param theClient
   * @param theVerb
   * @param theSet
   * @param theMetadataPrefix
   * @param maxNumberOfPagesToTest
   * @param recordsExpectedPerRequest
   * @param theThreadName
   * @param from
   * @param until
   * @param resumptionTokenANDMetadataPrefix
   * @throws IOException
   */
  public static void examineTGList(Client theClient, final String theVerb, final String theSet,
      final String theMetadataPrefix, final int maxNumberOfPagesToTest,
      final int recordsExpectedPerRequest, final String theThreadName, final String from,
      final String until, final boolean resumptionTokenANDMetadataPrefix) throws IOException {

    String threadName = tn(theThreadName);

    long startTime = System.currentTimeMillis();

    String testOccurance = "header";
    if (theVerb.equals(VERB_LIST_RECORDS)) {
      testOccurance = "record";
    }

    String query = "verb=" + theVerb + "&metadataPrefix=" + theMetadataPrefix;

    if (theSet != null) {
      query += "&set=" + theSet;
    }

    if (from != null && until != null) {
      query += "&from=" + from + "&until=" + until;
    }

    Response httpResponse = getOAIHttpResponse(theClient, threadName, query);
    int status = httpResponse.getStatus();

    int loopCount = 1;
    long timeRunning = System.currentTimeMillis() - startTime;
    System.out.println("\t" + threadName + "time: " + getDurationInSecs(timeRunning) + " (loop "
        + loopCount + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

    String responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    // Test resumption token tags.
    String restok = examineTGResumptionTokenTag(responseString, testOccurance, NO_TOKEN,
        recordsExpectedPerRequest, loopCount, threadName);

    // Test if the header identifiers are sound.
    if (theVerb.equals(VERB_LIST_RECORDS)) {
      examineTGHeader(responseString, recordsExpectedPerRequest);
    }

    // Test general metadata content (if verb is listRecords!), must go conform with the metadata
    // prefix setting.
    if (theVerb.equals(VERB_LIST_RECORDS)) {
      examineTGContent(responseString, theMetadataPrefix);
    }

    while (status == HttpStatus.SC_OK && !restok.equals(NO_TOKEN)) {
      loopCount += 1;
      synchronized (threadName) {
        if (maxNumberOfPagesToTest > 0 && loopCount > maxNumberOfPagesToTest) {
          System.out.println(
              "\t" + threadName + ">>> TESTING ONLY " + maxNumberOfPagesToTest + " pages!");
          break;
        }
        query = "verb=" + theVerb + "&resumptionToken=" + restok;
        if (resumptionTokenANDMetadataPrefix) {
          query += "&metadataPrefix=" + theMetadataPrefix;
        }
        httpResponse = getOAIHttpResponse(theClient, threadName, query);
        timeRunning = System.currentTimeMillis() - startTime;
        System.out.println("\t" + threadName + "time: " + getDurationInSecs(timeRunning) + " (loop "
            + loopCount + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

        // Test resumption token tags.
        responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());
        restok = examineTGResumptionTokenTag(responseString, testOccurance, restok,
            recordsExpectedPerRequest, loopCount, threadName);

        // Test if the header identifiers are sound.
        if (theVerb.equals(VERB_LIST_RECORDS)) {
          examineTGHeader(responseString, recordsExpectedPerRequest);
        }

        // Test general metadata content (if verb is listRecords!), must go conform with the
        // metadata prefix setting.
        if (theVerb.equals(VERB_LIST_RECORDS)) {
          examineTGContent(responseString, theMetadataPrefix);
        }
      }
    }

    // Only check for max loops, if we do have less loops and no resumption token, it will be fine,
    // too!
    if (!restok.equals(NO_TOKEN) && loopCount < maxNumberOfPagesToTest) {
      String message = threadName + ": Must have done " + maxNumberOfPagesToTest
          + " loops, but did only " + loopCount;
      assertTrue(message, false);
    }

    // Check for single loop, we need MORE!
    if (loopCount == 1) {
      String message = threadName + ": Must have more than one loop!";
      assertTrue(message, false);
    }

    System.out.println("\t" + threadName + OK);
  }

  /**
   * @param theResponse
   * @param recordOrHeader
   * @param oldtok
   * @param recordsExpectedPerRequest
   * @param loopCount
   * @param theThreadName
   * @return Resumption token string, if existing, -1 if either no <resumptionToken> tag is
   *         existing, tag is existing and has no token value.
   * @throws IOException
   */
  private static String examineTGResumptionTokenTag(final String theResponseString,
      final String recordOrHeader, final String oldtok, final int recordsExpectedPerRequest,
      final int loopCount, final String theThreadName) throws IOException {

    // Test for OAIPMH errors.
    if (theResponseString.contains("<error code=\"badArgument\">")) {
      String message = theThreadName + ": ERROR IN OAIPMH RESPONSE: " + theResponseString;
      System.out.println(message);
      throw new IOException(message);
    }

    // Count response objects at first.
    int recordCount = 0;
    int i = theResponseString.indexOf("<" + recordOrHeader + ">", 0);
    while (i != -1) {
      recordCount++;
      i++;
      i = theResponseString.indexOf("<" + recordOrHeader + ">", i);
    }

    System.out.println("\t" + theThreadName + recordOrHeader + "s: " + recordCount);

    // Check if token tag is existing.
    int tokStart = theResponseString.indexOf("<resumptionToken");
    int tokEnd = theResponseString.indexOf("</resumptionToken");

    if (tokStart == -1 && tokEnd == -1) {
      System.out.println("\t" + theThreadName + "token: no token");

      return NO_TOKEN;
    }

    String restokTmp = theResponseString.substring(tokStart, tokEnd);
    // Get token tag.
    String toktag = restokTmp.substring(0, restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "tokentag: " + toktag);

    // Get token.
    String restok = restokTmp.substring(restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "token: " + restok);

    // Check if old and new token are equal or not.
    boolean tokchanged = !oldtok.equals(restok);
    System.out.println("\t" + theThreadName + "tokchngd: " + tokchanged);

    // Get completeListSize and cursor.
    String sizeStr = toktag.substring(toktag.indexOf("completeListSize=\"") + 18);
    int size = Integer.parseInt(sizeStr.substring(0, sizeStr.indexOf("\"")));
    String cursorStr = toktag.substring(toktag.indexOf("cursor=\"") + 8);
    int cursor = Integer.parseInt(cursorStr.substring(0, cursorStr.indexOf("\"")));
    System.out.println("\t" + theThreadName + "size: " + size + " / " + cursor);

    // If token is provided, test cursor and element count!
    if (!restok.isEmpty()) {
      synchronized (OAIPMHUtilitiesOnline.class) {
        // Check <record> or <header> count, must be recordsExpectedPerRequest!
        if (recordCount != recordsExpectedPerRequest) {
          String message = recordOrHeader + " count mismatch, must be "
              + recordsExpectedPerRequest + " if token is provided, but is " + recordCount;
          assertTrue(message, false);
        }
        if (size <= recordsExpectedPerRequest) {
          String message = "completeListSize count mismatch, must be > "
              + recordsExpectedPerRequest + " if token is provided, but is " + size;
          assertTrue(message, false);
        }
        if (cursor != recordsExpectedPerRequest * loopCount) {
          String message = "cursor must be " + (loopCount * recordsExpectedPerRequest)
              + " in loop " + loopCount + ", but is " + cursor;
          assertTrue(message, false);
        }
      }
    }

    // If no token is provided, stop querying.
    else {
      // Check <record> count, must be completeListSize % recordsExpectedPerRequest.
      if (recordCount != size % recordsExpectedPerRequest) {
        String message = theThreadName + ": " + recordOrHeader + " count mismatch, should be "
            + size % recordsExpectedPerRequest + ", but is " + recordCount;
        assertTrue(message, false);
      }

      // No resumption token available in response.
      return NO_TOKEN;
    }

    return restok;
  }

  /**
   * @param theResponseString
   * @param theMetadataFormat
   */
  private static void examineTGContent(String theResponseString, String theMetadataFormat) {

    // Check for correct metadata content according to metadata prefix.
    if (theMetadataFormat.equals(OAI_DC_PREFIX)
        && !theResponseString.contains("metadataPrefix=\"" + OAI_DC_PREFIX + "\"")
        && !theResponseString.contains(EXPECTED_OAIDC_FORMAT_CONTENT)) {

      String message = OAI_DC_PREFIX + " needs to deliver content with schema: "
          + EXPECTED_OAIDC_FORMAT_CONTENT + "!";
      assertTrue(message, false);

    } else if (theMetadataFormat.equals(OAI_IDIOMMETS_PREFIX)
        && !theResponseString
            .contains("metadataPrefix=\"" + EXPECTED_IDIOMMETS_FORMAT_CONTENT + "\"")
        && !theResponseString.contains(EXPECTED_IDIOMMETS_FORMAT_CONTENT)) {

      String message = OAI_IDIOMMETS_PREFIX + " needs to deliver content with schema: "
          + EXPECTED_IDIOMMETS_FORMAT_CONTENT + "!";
      assertTrue(message, false);

    } else if (theMetadataFormat.equals(OAI_DATACITE_PREFIX)
        && !theResponseString.contains("metadataPrefix=\"" + OAI_DATACITE_PREFIX + "\"")
        && !theResponseString.contains(EXPECTED_DATACITE_FORMAT_CONTENT)) {

      String message = OAI_DATACITE_PREFIX + " needs to deliver content with schema: "
          + EXPECTED_DATACITE_FORMAT_CONTENT + "!";
      assertTrue(message, false);

    } else {
      // DEBUG ONLY!
      // System.out.println(theResponseString);
      // DEBUG ONLY!
    }
  }

  /**
   * <p>
   * Check for correct metadata content according to metadata prefix --> We do need to check if we
   * have different identifiers in the response.
   * </p>
   * 
   * @param theResponseString
   * @param recordsExpectedPerRequest
   */
  private static void examineTGHeader(String theResponseString, int recordsExpectedPerRequest) {
    boolean IDSUnique = examineTGHeaderIDs(theResponseString);
    if (IDSUnique) {
      String message = "Duplicate IDs in the response!";
      assertTrue(message, false);
    }
  }

  /**
   * <p>
   * Creates a hash set containing all the ID values from a certain tag. Used to count elements such
   * as <identifier> in OAI-PMH header list responses. We want to know here, if we have ID
   * duplicates in a response.
   * </p>
   * 
   * @param theResponseString
   * @return
   */
  private static boolean examineTGHeaderIDs(String theResponseString) {

    boolean result = false;
    HashSet<String> idHash = new HashSet<String>();

    int startID = 0;
    int endID = 0;
    String id = "";
    int count = 0;
    // Look for all the header ID tags.
    while (startID != -1) {
      count++;
      startID = theResponseString.indexOf(HEADER_ID_START_TAG, endID);
      endID = theResponseString.indexOf(HEADER_ID_END_TAG, startID);
      id = theResponseString.substring(startID + HEADER_ID_START_TAG.length(), endID);
      idHash.add(id);
    }
    // Remove "-1" ID again (and the respective counter).
    idHash.remove(id);
    count--;

    // Check for ID count: Hash size (unique IDs!) must be equal to the count of ID headers!
    if (count != idHash.size()) {
      String message =
          "We must have " + count + " unique IDs in the headers, but we got only " + idHash.size();
      assertTrue(message, false);
    }

    return result;
  }

  // **
  // DH used methods
  // **

  // FIXME

  // **
  // Commonly used (TG & DH) methods
  // **

  /**
   * @param theThreadName
   * @return
   */
  private static String tn(String theThreadName) {

    String result = theThreadName;

    if (theThreadName != null && !theThreadName.equals("") && !theThreadName.endsWith(" ")) {
      result = "[" + theThreadName + "] ";
    }

    return result;
  }

}
